from .entities import (
    Entity,
    Class,
    Individual,
    DataProperty,
    ObjectProperty,
    DataFieldType,
    Datatype,
    DatatypeType,
)
from .relation import TargetStatement, StatementType, RestrictionType, Relation
from .combined_relations import (
    CombinedRelation,
    CombinedObjectRelation,
    CombinedDataRelation,
)
from .source import DependencyStatement, ParsingError, Source
from .vocabulary import IdType, LabelSummary, VocabularySettings, Vocabulary
